/* find.c -- Windows95 IFS routines for FindFirstFile/FindNextFile
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include "mc.h"

#include <string.h>
#pragma intrinsic (memcpy, memset, strcmp, strlen, strcpy)


#pragma VxD_LOCKED_DATA_SEG
#pragma VxD_LOCKED_CODE_SEG

static struct hndlmisc hfFind = {
	IFS_VERSION,
	IFS_REVISION,
	NUM_HNDLMISC,

	ErrorStub,				/* FileSeek */
	FindFileClose,			/* Close */
	ErrorStub,				/* Commit */
	ErrorStub,				/* LockFile,*/
	ErrorStub,				/* FileDateTime, */

	ErrorStub,				/* NamedPipeRequest */
	ErrorStub,				/* NamedPipeHandleInfo, */
	EnumerateFind,			/* EnumerateHandle XXX -- should be handled*/
};


int SearchFile (pioreq pir)
{
	struct request *req = NULL;
	int res = ERROR_INVALID_FUNCTION;

	logmsg ("--searchfile--");

error:
	if (req) FreeRequest(req);
	logmsg ("--end of searchfile--");
	return pir->ir_error = res;
}

struct FindFile {
	ViceFid		ff_VFid;
	PREADDIR	ff_preaddir;
	DWORD		ff_attr;
	WORD		ff_pattern[MAX_PATH + 2];
	WORD		ff_thisname[LFNMAXNAMELEN +1];
	CTENTRY		ff_contable;
};

static int getnextmatch (struct FindFile *ff, _WIN32_FIND_DATA *wfd)
{
	DWORD matchflag = 0;
	DWORD hasdot = 0;
	DWORD excludeattr, mustattr;
	DWORD len, attr, type;
	ViceFid fid;
	struct coda_vattr vattr;
	int res = 0, i=1;

	if (ff->ff_attr & (FILE_FLAG_KEEP_CASE | FILE_FLAG_IS_LFN))
		matchflag |= UFLG_NT;
	else
		matchflag |= UFLG_DOS;
	if (ff->ff_attr & FILE_FLAG_WILDCARDS)
		matchflag |= UFLG_META;

	hasdot = (ff->ff_attr & FILE_FLAG_HAS_DOT) != 0;
	excludeattr = (~ff->ff_attr) & (FILE_ATTRIBUTE_HIDDEN | 
		FILE_ATTRIBUTE_SYSTEM | FILE_ATTRIBUTE_DIRECTORY);
	mustattr = ((ff->ff_attr & FILE_ATTRIBUTE_MUSTMATCH) >> 8);
	/* special case (thank you, Microsoft!) -- if LABEL attribute is
	   set in the "may-have" category, make it set in "must-have" because
	   that's the interpretation that Microsoft uses */
	if (ff->ff_attr & FILE_ATTRIBUTE_LABEL)
		mustattr |= FILE_ATTRIBUTE_LABEL;


	if (!(matchflag & UFLG_META))
		logucode (ff->ff_pattern);

	
	while ((res = readdir(ff->ff_preaddir, (PCHAR) ff->ff_thisname, &len)) 
			== 0) {
		if (len == 0) {
			/* end of directory */
			res = ERROR_NO_MORE_FILES;
			break;
		}
		((PCHAR) ff->ff_thisname)[len] = 0; 

		/* see if name matches */
		BCSToUni (wfd->cFileName, (PCHAR) ff->ff_thisname, len, BCS_OEM);
		wfd->cFileName[len] = 0;
		UniToUpper (wfd->cFileName, wfd->cFileName, len * 2);

		logucode (wfd->cFileName);
		logucode (ff->ff_pattern);

		if (!IFSMgr_MetaMatch(ff->ff_pattern, wfd->cFileName, matchflag))
			continue;

		/* OK, now test attributes */
		res = lookup(ff->ff_VFid, (PCHAR) ff->ff_thisname, &fid, &type, CLU_CASE_INSENSITIVE);

		if (strcmp("..", (PCHAR) ff->ff_thisname) == 0)
				continue;

		if (res) {
			printf ("lookup error %x", res);
			goto error;
		}

		attr = ((type == C_VDIR) ? FILE_ATTRIBUTE_DIRECTORY : 0);
		printf ("attr %x excl %x must %x", attr, excludeattr, mustattr);
		if ((attr & excludeattr) || 
			((attr & mustattr) != mustattr))
			continue;

		/* got it! */
		break;
	}
	if (!res) {

	//	PCHAR s = _HeapAllocate(13, 0);
	//	PCHAR l = _HeapAllocate(len,0);
		res = getattr(fid, &vattr);
		if (res) {
			printf ("getattr error %x", res);
			goto error;
		}
		
		memset ((PCHAR) wfd, 0, sizeof(_WIN32_FIND_DATA));
		BCSToUni (wfd->cFileName, (PCHAR) ff->ff_thisname, len, BCS_OEM);
		wfd->cFileName[len] = 0;
		wfd->nFileSizeHigh = 0;
		wfd->nFileSizeLow = (unsigned long) vattr.va_size;
		wfd->dwFileAttributes = attr;
		wfd->ftLastAccessTime = IFSMgr_NetToWin32Time(vattr.va_atime.ts_sec);
		wfd->ftLastWriteTime = IFSMgr_NetToWin32Time(vattr.va_mtime.ts_sec);
		wfd->ftCreationTime = IFSMgr_NetToWin32Time(vattr.va_ctime.ts_sec);
//		strcpy(l, (PCHAR) ff->ff_thisname);
//		convert_table_add(&ff->ff_contable, &s, l);
//		BCSToUni(wfd->cAlternateFileName, s, strlen(s), BCS_OEM);
//		wfd->cAlternateFileName[strlen(s)] = 0;
//		_HeapFree(s,0);
	//	_HeapFree(l,0);
	}
error:
	return res;
}

int FindFirstFile (pioreq pir)
{
	struct FindFile *ff = NULL;
	struct cnode *dircnode = NULL;

	int res = 0;

	logmsg ("--findfirstfile--");

	logucode (pir->ir_upath);

	print_ppath(pir->ir_ppath);
	res = open_namei(pir->ir_ppath, &dircnode);
	if (res) 
		return pir->ir_error = res;

	ff = _HeapAllocate(sizeof(struct FindFile), 0);
	if (!ff) {
		logmsg ("out of memory");
		res = ERROR_NOT_ENOUGH_MEMORY;
		goto error;
	}

//	convert_table_init(&ff->ff_contable);
	
	ff->ff_preaddir = NULL;
	res = opendir(dircnode->c_fid, &ff->ff_preaddir);
	if (res) {
		printf ("opendir failed %x", res);
		goto error;
	}

	/* copy the last path element into the pattern field -- 
	   here we exploit the fact that following the last path element is
	   a zero short, which we can use to delimit our pattern */

	memcpy ((PCHAR) ff->ff_pattern, 
			(PCHAR) (&IFSLastElement(pir->ir_ppath)->pe_unichars),
			IFSLastElement(pir->ir_ppath)->pe_length);

	/* upcase the pattern */
	UniToUpper (ff->ff_pattern, ff->ff_pattern, 
				IFSLastElement(pir->ir_ppath)->pe_length);

	ff->ff_attr = pir->ir_attr;
	ff->ff_VFid = dircnode->c_fid;

	logucode (ff->ff_pattern);
	printf ("specified attr %x", ff->ff_attr);

	res = getnextmatch(ff, (_WIN32_FIND_DATA *) pir->ir_data);
	if (res == ERROR_NO_MORE_FILES)
		res = ERROR_FILE_NOT_FOUND;

	if (!res) {
		pir->ir_fh = (PVOID) ff;
		printf("--findfirstfile ff handle %x", ff);
		SetHandleFunc (pir, FindNextFile, ErrorStub, &hfFind);
	}

error:

	closedir(ff->ff_preaddir);

	if (res) {

//		if (ff){		//not needed because FindFileClose does this
//			if (ff->ff_preaddir) _HeapFree(ff->ff_preaddir, 0);
//			_HeapFree(ff, 0);
//		}
		if (dircnode) iput (dircnode);
	}
	printf ("--end of findfirstfile: %x--", res);
	return pir->ir_error = res;
}

int FindNextFile (pioreq pir)
{
	struct FindFile *ff = (struct FindFile *) pir->ir_fh;
	int res = 0;

	logmsg ("--findnextfile--");
	printf("--findnextfile ff handle %x", ff);
	res = opendir(ff->ff_preaddir->VFid, &ff->ff_preaddir);
	if (res) {
		printf ("opendir failed %x", res);
		return pir->ir_error = res;
	}
	
	res = getnextmatch (ff, (_WIN32_FIND_DATA *) pir->ir_data);

	closedir(ff->ff_preaddir);
	printf ("--end of findnext: %x--", res);
	return pir->ir_error = res;
}


int FindFileClose (pioreq pir)
{
	struct FindFile *ff = (struct FindFile *) pir->ir_fh;
	int res =0;
	logmsg ("--findclose--");
	printf("--findfileclose pir handle %x", pir);
	printf("--findfileclose ff handle %x", ff);

	if (!res && ff){
		if (ff->ff_preaddir) _HeapFree(ff->ff_preaddir, 0);
//		convert_table_flush(&ff->ff_contable);
		_HeapFree(ff, 0);
	}

	printf ("--end of findclose: %x--", res);
	return pir->ir_error = res;
}

int EnumerateFind (pioreq pir)
{
	logmsg ("--enumeratefind--");
	logmsg ("--end of enumeratefind--");
	return pir->ir_error = ERROR_INVALID_FUNCTION;
}
