/* misc.c -- miscellaneous routines
   Copyright 1997-98 Michael Callahan
   Copyright 1999 Carnegie Mellon University
   This program is free software.  You may copy it according
   to the conditions of the GNU General Public License version 2.0
   or later; see the file COPYING in the source directory.
*/

#include <string.h>
#pragma intrinsic (strlen, strcpy)

#include "MC.h"
#include "vwin32wrap.h"
#include "vmmextra.h"

#pragma VxD_LOCKED_CODE_SEG
#pragma VxD_LOCKED_DATA_SEG

int errno2res (int errno)
{
	switch (errno) {
	case 0:  /* success */
		return 0;

	case ENOENT:
		return ERROR_FILE_NOT_FOUND;

	case EEXIST:
		return ERROR_FILE_EXISTS;

	case EPERM:
		return ERROR_ACCESS_DENIED;

	case EACCES:
		return ERROR_ACCESS_DENIED;

	case EISDIR:
		return ERROR_ACCESS_DENIED;  /* can do better? */

	case ENOTDIR:
		return ERROR_PATH_NOT_FOUND;

	case ENOTEMPTY:
		return ERROR_DIR_NOT_EMPTY;

	default:
		printf ("errno %x untranslated", errno);
		return ERROR_BAD_NET_RESP;
	}
}

char ucodebuf[400] = "foo";

void logucode (PWORD pu)
{
	PCHAR pc = ucodebuf;

	for ( ; *pu; pu++, pc++)
		if (*pu < 256)
			*pc = (char) *pu;
		else
			*pc = '?';
	*pc = '\0';
	logmsg (ucodebuf);
}


void logmsg (char *buffer)
{
	DWORD handle, action, size, nWritten;
	int err;

	err = R0_OpenCreateFile (FALSE, ACCESS_WRITEONLY | SHARE_DENYWRITE,
		0, (BYTE) ACTION_OPENALWAYS, (BYTE) R0_NO_CACHE, "C:\\VXD.LOG", &handle, &action);
	if (err)
		return;
	if (R0_GetFileSize (handle, &size))
		goto error;
	R0_WriteFile (FALSE, handle, strlen((const char *)buffer), size, (PBYTE) buffer, &nWritten);
	if (R0_GetFileSize (handle, &size))
		goto error;
	R0_WriteFile (FALSE, handle, 2, size, (const unsigned char *)"\r\n", &nWritten);
error:
	R0_CloseFile (handle);
}

struct request * NewRequest (int extrasize)
{
	struct request *req;

	req = (struct request *) _HeapAllocate (sizeof(struct request) + extrasize, HEAPZEROINIT);
	if (!req) {
		logmsg ("NewRequest: out of memory");
		return NULL;
	}
	req->age = 0;
	req->length = USER_LEVEL_REQUEST_SIZE + extrasize;
	return req;
}

void FreeRequest (struct request *req)
{
	if (req)
		_HeapFree(req, 0);
	else
		logmsg ("FreeRequest: NULL pointer");
}

struct cnode * NewCnode (void)
{
	struct cnode *cn;

	cn = (struct cnode *) _HeapAllocate (sizeof(struct cnode), HEAPZEROINIT);
	if (!cn) {
		logmsg ("NewCnode: out of memory");
		return NULL;
	}
	return cn;
}

void FreeCnode (struct cnode *cn)
{
	if (cn)
		_HeapFree(cn, 0);
	else
		logmsg ("FreeCnode: NULL pointer");
}

void fmt_dword (char *p, DWORD dw)
{
	int i;
	unsigned int j;

	for (i = 7; i >= 0; i--) {
		j = dw & 0xf;
		p[i] = "0123456789abcdef"[j];
		dw = dw >> 4;
	}
}

void print_pir (pioreq p)
{
	DWORD *pdw = (DWORD *) p;
	static char *m[] = {
		"xxxxxxxx  attr             xxxxxxxx  flags/user/sfn",
		"xxxxxxxx  pid              xxxxxxxx  ppath",
		"xxxxxxxx  aux1             xxxxxxxx  data",
		"xxxxxxxx  options/error    xxxxxxxx  rh",
		"xxxxxxxx  fh               xxxxxxxx  pos",
		"xxxxxxxx  aux2             xxxxxxxx  aux3",
		"xxxxxxxx  pev              xxxxxxxx  fsdata[0]"};
	int i;

	for (i = 0; i < 7; i++) {
		fmt_dword(m[i], *pdw++);
		fmt_dword(m[i] + 27, *pdw++);
		logmsg (m[i]);
	}
}

void printf(char *fmt, ...)
{
	static char buf[200] = "   ";
	static PVMMMUTEX printf_mtx = NULL;
	int *args, i;

	if (!printf_mtx) {
		printf_mtx = _CreateMutex(0, 0);
		if (!printf_mtx)
			logmsg ("INIT printf--cannot get MUTEX");
	}

	if (printf_mtx)
		_EnterMutex (printf_mtx, BLOCK_SVC_INTS | BLOCK_ENABLE_INTS);

	args = ((int *) &fmt) + 1;

	for (i = 0; *fmt && i < 180; fmt++, i++) {
		if ((*fmt != '%') || (*(fmt+1) != 'x')) 
			buf[i] = *fmt;
		else {
			fmt_dword (&buf[i], *args++);
			i += 7;
			fmt++;
		}
	}
	buf[i] = 0;
	logmsg (buf);
	if (printf_mtx)
		_LeaveMutex (printf_mtx);	
}

void print_ppath(ParsedPath* pp){
	PathElement *pe;	

	logmsg ("start print ppath");	
	pe = pp->pp_elements;	
//	while (pe != IFSLastElement(pp)){
//		logmsg ("test");
		logucode (pe->pe_unichars);
//		pe = IFSNextElement(pe);
//	}
	logmsg ("end print ppath");	
}

PUSHORT to_ucode(PCHAR name){
	PUSHORT  Uniname = _HeapAllocate(strlen(name)*2, 0);
	if (Uniname)
		BCSToUni(Uniname, name, strlen(name), BCS_OEM);
	return Uniname;
}
PCHAR to_BCS(PUSHORT name, int length){
	char *BCSname = _HeapAllocate(length+1, 0);
	if (length > MAX_PATH) return NULL;
	if (BCSname) {
		UniToBCS(BCSname, name, length, length, BCS_OEM);
		BCSname[length+1] = '\0';
	}
	return BCSname;
}
